# Pre-processing single cell data
# Author: John T Lafin
# Updated: 2024-09-17

print("Script begin at: ")
Sys.time()

# Setup -------------------------------------------------------------------

# Load packages
library(dplyr)
library(Seurat)
library(ggplot2)
library(intrinsicDimension)
library(clustree)
library(future)
library(future.apply)

# Future initialization
future_plan <- "multisession" # Can use multicore if not on Rstudio or Windows
n_workers <- 4
plan(future_plan, workers=n_workers)  
options(future.globals.maxSize = 8000 * 1024^2)
set.seed(42)

# Create output dir
outdir <- "data_modified/02_preprocess/"
dir.create(outdir, showWarnings = FALSE)

# Load data
seu <- readRDS("data_modified/01_initial_analysis/merged_object.rds")

# Clustering --------------------------------------------------------------

# Evaluate dimensionality
numPCs <- maxLikGlobalDimEst(seu@reductions$harmony@cell.embeddings, 
                             k = 10)$dim.est %>%
  round()

ElbowPlot(seu, ndims=50, reduction="harmony") + 
  geom_vline(xintercept=numPCs)
ggsave(paste0(outdir, "scree_plot.png"))

res <- seq(0.1,1,0.1)
seu <- RunUMAP(seu, 
                reduction = "harmony", 
                dims = 1:numPCs,
                metric = "euclidean") %>%
  FindNeighbors(reduction = "harmony",
                dims = 1:numPCs) %>%
  FindClusters(resolution = res)

# Ensure idents are in numerical order
for (i in res) {
  id <- paste0("RNA_snn_res.", i)
  top.level <- length(levels(seu@meta.data[[id]])) - 1
  seu@meta.data[[id]] <- factor(seu@meta.data[[id]], levels = seq(0, top.level, 1))
}

# Create clustree plot
clustree(seu, prefix = "RNA_snn_res.")
ggsave(paste0(outdir, "clustree.png"))

# Create UMAP plots at various resolutions
plotdir <- paste0(outdir, "umap/")
dir.create(plotdir, showWarnings = FALSE)

for (i in paste0("RNA_snn_res.", res)) {
  DimPlot(seu, group.by=i, label=TRUE, repel=TRUE)
  ggsave(paste0(plotdir, i, ".png"))
}


# Summary plots -----------------------------------------------------------

default.res <- "RNA_snn_res.0.4"
Idents(seu) <- default.res

# Fractional composition
cells.per.sample <- seu@meta.data %>%
  group_by(Sample, .data[[default.res]]) %>%
  summarize(count = n()) %>%
  group_by(.data[[default.res]]) %>%
  mutate(Fraction = count/sum(count))

ggplot(cells.per.sample, aes(.data[[default.res]], Fraction, fill = Sample)) + 
  geom_col() + 
  labs(title = "Fraction composition",
       fill = "")
ggsave(paste0(outdir, "frac_comp.png"))

# Doublets per cluster
dblts <- seu@meta.data %>%
  group_by(.data[[default.res]], scDblFinder.class) %>%
  summarize(count = n()) %>%
  group_by(.data[[default.res]]) %>%
  mutate(fraction = count/sum(count))

ggplot(dblts, aes(.data[[default.res]], fraction, fill = scDblFinder.class)) +
  geom_col() + 
  geom_hline(yintercept = 0.5, linetype = "dashed")
ggsave(paste0(outdir, "dblts.png"))

# Lineage markers
genes <- strsplit("EPCAM KRT5 KRT14 MSMB SEMG1 SELE PECAM1 PTPRC SRGN C1QA DCN LUM DES ACTA1 ACTG2 CRYAB CDH19",
                  split = " ")[[1]]
DotPlot(seu, features = genes, cluster.idents = TRUE) + 
  theme(axis.text.x = element_text(angle = 45, hjust = 1))
ggsave(paste0(outdir, "lineage_marker_dotplot.png"))

# Marker genes ------------------------------------------------------------

markerdir <- paste0(outdir, "markers/")
dir.create(markerdir, showWarnings=FALSE)

for (i in paste0("RNA_snn_res.", res)) {
  Idents(seu) <- i
  m <- FindAllMarkers(seu, only.pos=TRUE)
  write.csv(m, paste0(markerdir, i, ".csv"))
}

# Wrap up -----------------------------------------------------------------

# Save processed data
saveRDS(seu, paste0(outdir, "processed_object.rds"))
capture.output(sessionInfo(), file = paste0(outdir, "sessionInfo.txt"))
