# Annotation and stress score filtering
# Author: John T Lafin
# Updated: 2024-10-16

print("Script begin at: ")
Sys.time()

# Setup -------------------------------------------------------------------

# Load packages
library(dplyr)
library(Seurat)
library(intrinsicDimension)
library(ggplot2)
library(harmony)
library(future)
library(future.apply)

# Future initialization
future_plan <- "multisession" # Can use multicore if not on Rstudio or Windows
n_workers <- 4
plan(future_plan, workers=n_workers)  
options(future.globals.maxSize = 8000 * 1024^2)
set.seed(42)

# Create output dir
outdir <- "data_modified/04_annotation/"
dir.create(outdir, showWarnings = FALSE)

# Create pre-processing function
preprocess <- function(x, 
                       cluster.res = seq(0.1,1.5,0.1), 
                       default.res = "RNA_snn_res.0.4", 
                       batch.var = "Sample"){
  x <- x %>%
    NormalizeData() %>%
    FindVariableFeatures() %>%
    ScaleData() %>%
    RunPCA() %>%
    RunHarmony(group.by.vars = batch.var)
  
  numPCs <- maxLikGlobalDimEst(x@reductions$harmony@cell.embeddings, 
                               k = 10)$dim.est %>%
    round()
  x <- x %>%
    RunUMAP(dims = 1:numPCs,
            reduction = "harmony",
            metric = "euclidean",
            min.dist = 0.1) %>%
    FindNeighbors(reduction = "harmony",
                  dims = 1:numPCs) %>%
    FindClusters(resolution = cluster.res) %>%
    SetIdent(value = default.res)
  return(x)
}

# Load data ---------------------------------------------------------------

lin.list <- list()
lins <- list.files('data_modified/03_lineage_filtering/objects/') %>%
  sub(".rds", "", .)
for (lin in lins) {
  lin.list[[lin]] <- readRDS(paste0("data_modified/03_lineage_filtering/objects/", lin, ".rds"))
}

# Annotation --------------------------------------------------------------

annotdir <- paste0(outdir, "annotations_first_pass/")
dir.create(annotdir, showWarnings=FALSE)

## Create list of labels
annot.list <- list()
annot.list$Endothelia <- c("venous endothelia",
                           "venous endothelia",
                           "capillary endothelia",
                           "capillary endothelia",
                           "arterial endothelia",
                           "lymphatic endothelia")
annot.list$Epithelia <- c("transitional basal epithelia of urethra",
                          "basal epithelia of prostate",
                          "transitional basal epithelia of urethra",
                          "squamous intermediate epithelia of urethra",
                          "basal epithelia of prostate",
                          "transitional intermediate epithelia of urethra",
                          "squamous intermediate epithelia of urethra",
                          "luminal epithelia of prostate",
                          "transitional luminal epithelia of urethra",
                          "transitional intermediate epithelia of urethra",
                          "transitional luminal epithelia of urethra",
                          "transitional basal (JUN/FOS hi) epithelia of urethra",
                          "transitional basal (JUN/FOS hi) epithelia of urethra",
                          "squamous intermediate epithelia of urethra",
                          "squamous intermediate epithelia of urethra",
                          "basal epithelia of prostate",
                          "squamous basal epithelia of urethra",
                          "basal epithelia of prostate",
                          "proliferative cells",
                          "squamous luminal epithelia of urethra",
                          "transitional basal (JUN/FOS hi) epithelia of urethra",
                          "transitional basal epithelia of urethra",
                          "basal epithelia of prostate")
annot.list$Leukocyte <- c("T cell",
                          "macrophage",
                          "mast cell",
                          "T cell",
                          "T cell",
                          "NK cell",
                          "proliferating leukocyte",
                          "B cell",
                          "dendritic cell",
                          "plasma cell")
annot.list$Stroma <- c("pericyte",
                       "interstitial fibroblast of urethra",
                       "smooth muscle cell",
                       "interstitial fibroblast of vagina",
                       "smooth muscle cell",
                       "smooth muscle cell",
                       "periepithelial fibroblast of vagina",
                       "vascular smooth muscle",
                       "smooth muscle cell",
                       "periepithelial fibroblast of urethra",
                       "stressed cell",
                       "pericyte",
                       "Schwann cell",
                       "smooth muscle cell",
                       "endoneurial fibroblast",
                       "pericyte",
                       "Schwann cell")

## Set epithelia and stroma to higher resolutions
Idents(lin.list$Epithelia) <- "RNA_snn_res.1.5"
Idents(lin.list$Stroma) <- "RNA_snn_res.1"

## Add the labels
for (lin in lins) {
  names(annot.list[[lin]]) <- levels(lin.list[[lin]])
  lin.list[[lin]] <- RenameIdents(lin.list[[lin]], annot.list[[lin]]) %>%
    StashIdent("cellannotation")
}

## Add neuroendocrine cells
NE.markers <- strsplit("CHGA GRP CALCA SCG2 TPH1 ASCL1 SCG5 PHGR1 MIAT MAOB", split = " ")
lin.list$Epithelia <- AddModuleScore(lin.list$Epithelia, 
                       features = NE.markers,
                       name = "NE.score")
lin.list$Epithelia@meta.data <- rename(lin.list$Epithelia@meta.data, NE.score = NE.score1)
VlnPlot(lin.list$Epithelia, "NE.score", group.by = "orig.ident") + 
  NoLegend() + 
  geom_hline(yintercept=0.3, linetype="dashed")
ggsave(paste0(annotdir, "NE_score_cutoff.png"))

lin.list$Epithelia@meta.data <- lin.list$Epithelia@meta.data %>%
  mutate(cellannotation = if_else(NE.score > 0.3, "neuroendocrine epithelia", cellannotation))

## Add ionocytes
VlnPlot(lin.list$Epithelia, "FOXI1", group.by = "orig.ident") + 
  NoLegend() + 
  geom_hline(yintercept=0.3, linetype="dashed")
ggsave(paste0(annotdir, "FOXI1_score_cutoff.png"))

ionocyte.barcodes <- WhichCells(lin.list$Epithelia, expression = FOXI1 > 0.3)
lin.list$Epithelia@meta.data[ionocyte.barcodes, "cellannotation"] <- "ionocyte"

## Split periepithelial fibroblasts by sex
fib_rename <- WhichCells(lin.list$Stroma, 
                         idents="periepithelial fibroblast of vagina", 
                         expression = Sex == "male")
levels(lin.list$Stroma$cellannotation) <- c(levels(lin.list$Stroma$cellannotation), "periepithelial fibroblast of prostate")
lin.list$Stroma@meta.data[fib_rename, "cellannotation"] <- "periepithelial fibroblast of prostate"

## Split interstitial fibroblasts by sex
fib_rename <- WhichCells(lin.list$Stroma, 
                         idents=c("interstitial fibroblast of vagina", "interstitial fibroblast of urethra"), 
                         expression = Sex == "male")
levels(lin.list$Stroma$cellannotation) <- c(levels(lin.list$Stroma$cellannotation), "interstitial fibroblast of prostate")
lin.list$Stroma@meta.data[fib_rename, "cellannotation"] <- "interstitial fibroblast of prostate"


## Reset idents
Idents(lin.list$Epithelia) <- "cellannotation"
Idents(lin.list$Stroma) <- "cellannotation"

## Save labeled UMAP plots
for (lin in lins) {
  DimPlot(lin.list[[lin]])
  ggsave(paste0(annotdir, lin,"_umap.png"))
}

# Stress score filtering --------------------------------------------------

stressdir <- paste0(outdir, "stress/")
dir.create(stressdir, showWarnings = FALSE)

## Produce violin plots of stress scores
lin.list <- lapply(lin.list, SetIdent, value = "RNA_snn_res.1")

### Check epithelia at higher resolution to capture other populations
Idents(lin.list$Epithelia) <- "RNA_snn_res.1.5"

for (lin in names(lin.list)) {
  VlnPlot(lin.list[[lin]], "Stress.score", pt.size=0) +
    NoLegend()
  ggsave(paste0(stressdir, lin, "_pre-filter.png"))
}

## Define stressed clusters based on violin plots
stressed.cells <- list()
stressed.cells$Stroma <- c(2,10,15)
stressed.cells$Endothelia <- c(4,10)
stressed.cells$Leukocyte <- c(4,9)

## Remove stressed clusters and reprocess
for (lin in names(lin.list)) {
  if (!(lin %in% names(stressed.cells))) {
    Idents(lin.list[[lin]]) <- "cellannotation"
    next
  }
  lin.list[[lin]] <- subset(lin.list[[lin]], 
                            idents = stressed.cells[[lin]],
                            invert = TRUE)
  lin.list[[lin]] <- preprocess(lin.list[[lin]], default.res="cellannotation")
}

## Save violin plots after first round of filtering
for (lin in names(lin.list)) {
  VlnPlot(lin.list[[lin]], "Stress.score", pt.size=0, group.by="RNA_snn_res.1") +
    NoLegend()
  ggsave(paste0(stressdir, lin, "_first_filter.png"))
}

## Second round of stressed cell removal in stroma
Idents(lin.list$Stroma) <- "RNA_snn_res.1"
lin.list$Stroma <- subset(lin.list$Stroma,
                          idents = 8,
                          invert = TRUE) %>%
  preprocess(default.res="cellannotation")

## Save violin plot
VlnPlot(lin.list$Stroma, "Stress.score", pt.size=0, group.by="RNA_snn_res.1") +
  NoLegend()
ggsave(paste0(stressdir, lin, "_second_filter.png"))

## Reannotate endothelia
lin.list$Endothelia@meta.data <- lin.list$Endothelia@meta.data %>% 
  mutate(cellannotation = if_else(RNA_snn_res.1 == 7, "venous endothelia", cellannotation))
Idents(lin.list$Endothelia) <- "cellannotation"

## Save labeled UMAP plots
for (lin in lins) {
  DimPlot(lin.list[[lin]])
  ggsave(paste0(annotdir, lin,"_umap-2.png"))
}

# Merge -------------------------------------------------------------------

seu <- merge(lin.list[[1]], lin.list[2:length(lin.list)])
plan(sequential)
seu <- preprocess(seu, default.res="cellannotation")

## Save UMAP plot
DimPlot(seu, label=TRUE, repel=TRUE, label.size=3) + 
  NoLegend()
ggsave(paste0(annotdir, "global_umap.png"))

# Save outputs ------------------------------------------------------------

objdir <- paste0(outdir, "objects/")
dir.create(objdir, showWarnings=FALSE)

saveRDS(seu, paste0(objdir, "global.rds"))
for (lin in names(lin.list)){
  saveRDS(lin.list[[lin]], paste0(objdir, lin, ".rds"))
}

## Save session info
capture.output(sessionInfo(), file = paste0(outdir, "session_info.txt"))
