# Calculate contamination fraction from cellbender output
# As per: https://github.com/broadinstitute/CellBender/issues/181
# Author: John T Lafin
# Updated: 2024-08-06


# Load modules
import csv
import numpy as np
import pandas as pd
import scanpy as sc
from cellbender.remove_background.downstream import load_anndata_from_input_and_output
from scipy import optimize
from scipy.stats import gaussian_kde
import matplotlib.pyplot as plt
import seaborn as sns

# Load list of files
input_tsv = 'sample_sheets/contamination_fraction.tsv'
df = pd.read_csv(input_tsv, sep='\t')

# Calculate rho (contamination fraction) per cell for each sample
rho = {}
for idx, row in df.iterrows():
    adata = load_anndata_from_input_and_output(
        input_file=row['cb_input'],
        output_file=row['cb_output'],
        input_layer_key='raw'
    )
    rho[row['sample']] = (adata.obs['n_raw'] - adata.obs['n_cellbender']) / adata.obs['n_raw']

# Find summary rho for each sample
rho_summary = {}
for key in rho:
    kernel = gaussian_kde(rho[key])
    rho_summary[key] = optimize.minimize_scalar(lambda x: -kernel(x), method="bounded", bounds=(-0.5,0.9)).x

# Save contamination fraction estimates
with open('data_raw/cellbender/contamination_estimates.csv', 'w') as f:
    writer = csv.writer(f)
    writer.writerow(['Sample','contamination_fraction'])
    for k, v in rho_summary.items():
        writer.writerow([k, v.item()])

# Plot contamination fraction and save
for k in rho:
    plotfile = 'data_raw/cellbender/' + k + '/contamination_estimate.png'
    fig, ax = plt.subplots(1,1)
    sns.kdeplot(rho[k], ax=ax).set_title(label=k)
    ax.axvline(rho_summary[k], linestyle='dashed', color='red')
    fig.savefig(plotfile)
